/*
** Copyright (C) Red Hat, Inc. 2005-2012
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING. If not, write to the
** Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
** MA 02139, USA.
*/

/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */

#include "XML.h"
#include "Mutex.h"
#include "utils.h"
#include "File.h"

#include <libxml/parser.h>
#include <libxml/tree.h>

#include <algorithm>

//#include <iostream>
using namespace std;

static String escape_chars(const String&);
static String invert_chars(const String&);

XMLObject::XMLObject(const String& elem_name) :
	_tag(elem_name)
{
	_node_type = XML_ELEMENT_NODE;
}

XMLObject::~XMLObject()
{}


// Compare attributes against a different XMLObject
// Make a copy of the right side's attrs and remove
// any duplicates found.
//
// If the list is nonempty when we're done, the lists
// differ.
bool
XMLObject::compare_attrs(const XMLObject& other) const
{
	const std::map<String, String>& left = attrs();
	const std::map<String, String>& right = other.attrs();

	if (left.size() != right.size())
		return false;

	for (map<String, String>::const_iterator x = left.begin();
		x != left.end(); x++) {

		map<String, String>::const_iterator y = right.find(x->first);

		/* not present */
		if (y == right.end())
			return false;

		if (x->second != y->second)
			return false;
	}

	return true;
}

bool
XMLObject::operator== (const XMLObject& obj) const
{
	if (tag() != obj.tag()) {
		return false;
	}

	if (compare_attrs(obj) != true) {
		return false;
	}

	if (children() != obj.children())
		return false;
	return true;
}

bool
XMLObject::operator!= (const XMLObject& obj) const
{
	return !(operator==(obj));
}

bool
XMLObject::remove_child(const XMLObject& child)
{
	list<XMLObject>::iterator iter = find(_kids.begin(), _kids.end(), child);
	if (iter == _kids.end())
		return false;
	else {
		_kids.erase(iter);
		return true;
	}
}

void
XMLObject::generate_xml(String& xml, const String& indent) const
{
	if (_node_type == XML_ELEMENT_NODE) {
		xml += indent + "<" + _tag;
		for (map<String, String>::const_iterator
			iter = attrs().begin() ;
			iter != attrs().end() ;
			iter++)
		{
			const String& name = iter->first;
			const String value = escape_chars(iter->second);
			xml += " " + name + "=\"" + value + "\"";
		}

		if (children().empty())
			xml += "/>\n";
		else {
			xml += ">\n";

			for (list<XMLObject>::const_iterator
				 iter = children().begin() ;
				 iter != children().end() ;
				 iter++)
			{
				iter->generate_xml(xml, indent + "\t");
			}
			xml += indent + "</" + _tag + ">\n";
		}
	} else if (_node_type == XML_COMMENT_NODE) {
		xml += indent + "<!-- " + _text + " -->\n";
	} else if (_node_type == XML_TEXT_NODE) {
		xml += _text;
	}
}

// *** GLOBAL FUNCTIONS ***

static void
_parseXML(XMLObject& parent, xmlNode* children)
{
	for (xmlNode *curr_node = children; curr_node ; curr_node = curr_node->next)
	{
		if (curr_node->type == XML_ELEMENT_NODE) {
			XMLObject me((const char *) curr_node->name);

			if (curr_node->nsDef) {
				const struct _xmlNs *cur_def;

				for (cur_def = curr_node->nsDef ;
					 cur_def != NULL ;
					 cur_def = cur_def->next)
				{
					if (cur_def->prefix) {
						me.set_attr("xmlns:" +
									String((const char *) cur_def->prefix),
							String((const char *) cur_def->href));
					} else {
						me.set_attr("xmlns",
							String((const char *) cur_def->href));
					}
				}
			}

			// attrs
			for (xmlAttr* curr_attr = curr_node->properties ;
					curr_attr ;
					curr_attr = curr_attr->next)
			{
				if (curr_attr->type == XML_ATTRIBUTE_NODE) {
					const xmlChar *name = curr_attr->name;
					const xmlChar *value = xmlGetProp(curr_node, name);
					String prefix = "";

					if (!value)
						throw String("xmlGetProp() returned NULL");

					if (curr_attr->ns)
						prefix = (const char *) curr_attr->ns->prefix + String(":");

					try {
						const String name_str = prefix + String((const char *) name);
						const String value_str =
							invert_chars((const char *) value);

						me.set_attr(name_str, value_str);
						xmlFree((void *) value);
					} catch ( ... ) {
						xmlFree((void *) value);
						throw;
					}
				}
			}

			// kids
			_parseXML(me, curr_node->children);
			parent.add_child(me);
		} else if ((curr_node->type == XML_COMMENT_NODE || curr_node->type == XML_TEXT_NODE)
					&& !xmlIsBlankNode(curr_node))
		{
			xmlChar *node_content = xmlNodeGetContent(curr_node);

			if (node_content) {
				char *tp = (char *) node_content;
				size_t content_len = strlen(tp);
				char *text_start;
				char *text_end;

				if (content_len > 0)
					text_end = &tp[content_len - 1];
				else
					text_end = tp;

				while (isspace(*tp))
					tp++;
				text_start = tp;

				while (text_end > text_start && isspace(*text_end))
					text_end--;
				*(text_end + 1) = '\0';

				String text((const char *) text_start);
				xmlFree((void *) node_content);

				XMLObject me("DUMMY");
				me.set_node_type(curr_node->type);
				me.set_text(text);
				parent.add_child(me);
			}
		}
	}
}

XMLObject
parseXML(const String& xml)
{
	static bool initialized = false;

	if (!initialized) {
		LIBXML_TEST_VERSION;
		initialized = true;
	}

	xmlDoc *doc = xmlReadMemory(xml.c_str(),
					xml.size(),
					"noname.xml",
					NULL,
					XML_PARSE_NONET | XML_PARSE_NOERROR | XML_PARSE_NOWARNING);

	if (!doc)
		throw String("parseXML(): couldn't parse xml");

	XMLObject root("if you see this, something wrong happened");

	try {
		_parseXML(root, xmlDocGetRootElement(doc));
		xmlFreeDoc(doc);
		return *(root.children().begin());
	} catch ( ... ) {
		xmlFreeDoc(doc);
		throw String("parseXML(): low memory");
	}
}


String
generateXML(const XMLObject& obj)
{
	String xml("<?xml version=\"1.0\"?>\n");
	obj.generate_xml(xml, "");

	// verify xml
	xmlDoc* doc = xmlReadMemory(xml.c_str(),
					xml.size(),
					"noname.xml",
					NULL,
					XML_PARSE_NONET | XML_PARSE_NOERROR | XML_PARSE_NOWARNING);
	if (!doc) {
		//cout << xml << endl;
		throw String("generateXML(): internal error");
	}
	xmlFreeDoc(doc);

	return xml;
}

XMLObject
readXML(const String& filename)
{
	return parseXML(File::open(filename));
}

String
escape_chars(const String& str)
{
	const String amp_repl("______AMP_REPLACEMENT_XML_KOJIKOJIKOJIKO______");
	const String lt_repl("______LT_REPLACEMENT_XML_KOJIKOJIKOJIKO______");
	const String gt_repl("______GT_REPLACEMENT_XML_KOJIKOJIKOJIKO______");
	const String apos_repl("______APOS_REPLACEMENT_XML_KOJIKOJIKOJIKO______");
	const String quot_repl("______QUOT_REPLACEMENT_XML_KOJIKOJIKOJIKO______");

	String ret = utils::replace("&amp;", amp_repl, str);
	ret = utils::replace("&lt;", lt_repl, ret);
	ret = utils::replace("&gt;", gt_repl, ret);
	ret = utils::replace("&apos;", apos_repl, ret);
	ret = utils::replace("&quot;", quot_repl, ret);

	ret = utils::replace("&", "&amp;", ret);
	ret = utils::replace("<", "&lt;", ret);
	ret = utils::replace(">", "&gt;", ret);
	ret = utils::replace("'", "&apos;", ret);
	ret = utils::replace("\"", "&quot;", ret);

	ret = utils::replace(amp_repl, "&amp;", ret);
	ret = utils::replace(lt_repl, "&lt;", ret);
	ret = utils::replace(gt_repl, "&gt;", ret);
	ret = utils::replace(apos_repl, "&apos;", ret);
	ret = utils::replace(quot_repl, "&quot;", ret);

	return ret;
}

String
invert_chars(const String& str)
{
	String ret = utils::replace("&amp;", "&", str);
	ret = utils::replace("&lt;", "<", ret);
	ret = utils::replace("&gt;", ">", ret);
	ret = utils::replace("&apos;", "'", ret);
	ret = utils::replace("&quot;", "\"", ret);
	return ret;
}
