/*
** Copyright (C) Red Hat, Inc. 2005-2009
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING. If not, write to the
** Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
** MA 02139, USA.
*/

/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */

#include "executils.h"
#include "Logger.h"
#include "Time.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/poll.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <syslog.h>

static void read_data(struct pollfd& poll_info, bool& fd_closed, String& data);
static void close_fd(int fd);

// NOTE: Execution of multi-process programs "parent exits, child continues"
//       requires caution (see "Since bz564490" comment/consult executils.h).
int
execute(const String& path,
		const std::vector<String>& args,
		String& out,
		String& err,
		int& status,
		int timeout)
{
	if (access(path.c_str(), X_OK))
		return 1;

	out = err = "";

	int _stdout_pipe[2];
	int _stderr_pipe[2];

	/*
	** Prepare argv before forking in case we were called from a thread.
	*/
	unsigned int size = args.size() + 2;
	char **argv = (char **) malloc(sizeof(char *) * size);
	if (argv == NULL)
		return 3;

	String comm(path);

	// get status
	try {
		argv[0] = (char *) path.c_str();
		for (unsigned int i = 0 ; i < args.size() ; i++) {
			argv[i + 1] = (char *) args[i].c_str();
			comm += " " + args[i];
		}
		argv[size - 1] = NULL;
	} catch (...) {
		free(argv);
		return 3;
	}

	if (pipe(_stdout_pipe) == -1) {
		free(argv);
		return 2;
	}

	if (pipe(_stderr_pipe) == -1) {
		free(argv);
		close_fd(_stdout_pipe[0]);
		close_fd(_stdout_pipe[1]);
		return 2;
	}

	setenv("LANG", "C", 1);
	setenv("LC_ALL", "C", 1);

	int pid = fork();
	if (pid == -1) {
		free(argv);
		close_fd(_stdout_pipe[0]);
		close_fd(_stdout_pipe[1]);
		close_fd(_stderr_pipe[0]);
		close_fd(_stderr_pipe[1]);
		return 3;
	}

	if (pid == 0) {
		/* child */

		close_fd(1);
		close_fd(_stdout_pipe[0]);
		dup2(_stdout_pipe[1], 1);
		close_fd(_stdout_pipe[1]);

		close_fd(2);
		close_fd(_stderr_pipe[0]);
		dup2(_stderr_pipe[1], 2);
		close_fd(_stderr_pipe[1]);

		int devnull = open("/dev/null", O_RDWR);
		if (devnull == -1) {
			fprintf(stderr, "ricci::execute(): Can't open /dev/null: %s\n",
				strerror(errno));
			_exit(1);
		}

		close_fd(0);
		dup2(devnull, 0);
		close_fd(devnull);

		// close open fds
		for (unsigned int i = 3 ; i < __FD_SETSIZE ; i++)
			close_fd(i);

		// restore signals
		for (int x = 1 ; x < _NSIG ; x++)
			signal(x, SIG_DFL);

		sigset_t set;
		sigfillset(&set);
		sigprocmask(SIG_UNBLOCK, &set, NULL);

		openlog("ricci", LOG_PID | LOG_NDELAY, LOG_DEBUG);
		syslog(LOG_NOTICE, "Executing '%s'", comm.c_str());
		closelog();

		execv(path.c_str(), argv);
		_exit(1);
	}

	/* parent */
	unsigned int time_beg = time_mil();
	unsigned int time_to_kill = time_beg + timeout;
	bool out_closed = false, err_closed = false;

	free(argv);
	close_fd(_stdout_pipe[1]);
	close_fd(_stderr_pipe[1]);

	while (true) {
		// kill child if timeout elapsed
		if (timeout >= 0 && time_mil() > time_to_kill) {
			kill(pid, SIGTERM);
			sleep(1);
			kill(pid, SIGKILL);
		}

		// prepare poll structs
		struct pollfd poll_data[2];
		int s = 0;
		if (!out_closed) {
			poll_data[s].fd = _stdout_pipe[0];
			poll_data[s].events = POLLIN;
			poll_data[s].revents = 0;
			s += 1;
		}

		if (!err_closed) {
			poll_data[s].fd = _stderr_pipe[0];
			poll_data[s].events = POLLIN;
			poll_data[s].revents = 0;
			s += 1;
		}

		if (s == 0)
			break;

		// wait for events
		int ret = poll(poll_data, s, 500);
		if (ret == 0) {
			// Since bz564490, we no longer rely on "pipes closed iff exit status ready"
			// (easy counterexample: parent in multi-process program exits while its
			// child keeps the descriptors unclosed; FD_CLOEXEC is not a remedy) and
			// thus we keep watching here whether the main process of executed program
			// has exited and once it has, we are done (no waiting for any child/ren).
			//
			// Rationale 1 behind taking care of the main process only:
			// a. common case -- exit on behalf of main process (single/multi-process)
			// b. multi-process program with child surviving main process
			//    -- rare case, we prefer exit~returned exit status (as with old virsh)
			//
			// Rationale 2 behind taking the test right here:
			// This branch guarantees that we don't miss any data yet to be read
			// (POLLIN -> has to read first even if the process has already exited);
			// POLLERR|POLLHUP|POLLNVAL cases for particular descriptor mean a half-way
			// pass the infloop where another, blocking, waitpid takes care, anyway.
			//
			// Note 1: Placing the test here causes two issues (solving of which is
			// not worth the code complexity growth or even possible code break):
			// a. the captured stdout/stderr may contain data produced by the child
			//    as per Rationale 1b. generated even AFTER the point parent exited
			//    (more writers in such case is strange, still parent data not missed)
			// b. child as per Rationale 1b. could prevent this branch to be taken by
			//    consistently generating data on stdin/stderr (even stranger scenario).
			//
			// Note 2: Preferred scenario as per Rationale 1b. will cause issues in
			// case child process, survived its parent, will try to use stdout/stderr
			// as these were initially (before local exec) turned into write ends of
			// respective pipes -- child's write will generate SIGPIPE signal
			// and further run may fail (depending on the child's robustness).
			int status;
			if (waitpid(pid, &status, WNOHANG) > 0 && WIFEXITED(status))
				break;
			continue;
		} else if (ret == -1) {
			if (errno == EINTR)
				continue;
			else {
				if (!out_closed)
					close_fd(_stdout_pipe[0]);
				if (!err_closed)
					close_fd(_stderr_pipe[0]);
				return 4;
			}
		}

		// process events
		for (int i = 0 ; i < s ; i++) {
			struct pollfd& poll_info = poll_data[i];

			if (poll_info.fd == _stdout_pipe[0])
				read_data(poll_info, out_closed, out);
			if (poll_info.fd == _stderr_pipe[0])
				read_data(poll_info, err_closed, err);
		}
	}

	int ret;
	do {
		ret = waitpid(pid, &status, 0);
	} while ((ret < 0) && (errno == EINTR));

	if (WIFEXITED(status)) {
		//log("executed \"" + comm + "\" in " + (time_mil() - time_beg) + " milliseconds", LogExecute);
		status = WEXITSTATUS(status);
		return 0;
	} else if (WIFSIGNALED(status)) {
		//log("\"" + comm + "\" killed after " + (time_mil() - time_beg) + " milliseconds", LogExecute);
		return 5;
	} else {
		//log("unknown cause of \"" + comm + "\"'s exit after " + (time_mil() - time_beg) + " milliseconds", LogExecute);
		return 6;
	}
}

void
read_data(struct pollfd& poll_info, bool& fd_closed, String& data)
{
	int fd = poll_info.fd;

	if (poll_info.revents & POLLIN) {
		try {
			char data_in[4096];
			int ret = read(fd, data_in, sizeof(data_in));
			if (ret < 0)
				// if it was errno==EINTR, POLLIN event will be kept, otherwise
				// presumably cleared and problem indicated with POLLERR|POLNVAL
				return;
			if (ret == 0) {
				// EOF;  it may be (without guarantees) accompanied with POLLHUP event
				// upon current/next poll (see [1]), but handle this safely in any case
				// [1] http://www.greenend.org.uk/rjk/2001/06/poll.html
				close_fd(fd);
				fd_closed = true;
				return;
			}
			data.append(data_in, ret);
		} catch ( ... ) {
			close_fd(fd);
			fd_closed = true;
		}
		// At this point, we have successfully read some data.
		// But in POLLIN|POLLHUP case, we cannot be sure whether we have already
		// read all:
		// a. ret == sizeof(data_in)
		//    -> trivially, we have no idea whether we read all yet, need to continue
		// b. 0 < ret < sizeof(data_in)
		//    -> as per READ(3P) man page, it could have been the fact that current
		//       read() was interrupted by signal (e.g., inevitable SIGCHLD) or
		//       simply there were less bytes immediately available (not probable
		//       in this context due to temporal order) that caused less bytes read
		//       than was ready in the pipe (i.e., one cannot put equal sign
		//       between "not a full buffer read" and "certain EOF ahead")
		// thus in such case, we have to return right now and let subsequent poll
		// or read (see comment for "ret == 0" branch above) decide
		// (POLLIN will remain set if we have still something to read; apparently,
		// POLLHUP will remain set until we close the descriptor).
		//
		// This way, we can be sure that we do not miss any data due to finishing
		// with POLLHUP handling (below) prematurely.
		// If we ignore POLLERR|POLLNVAL this time (unlikely), the right branch
		// should be taken the next iteration (with POLLIN presumably cleared).
		if (poll_info.revents & POLLHUP)
			return;
	}

	if (poll_info.revents & (POLLERR | POLLHUP | POLLNVAL)) {
		close_fd(fd);
		fd_closed = true;
		return;
	}
}

void
close_fd(int fd)
{
	int e;
	do {
		e = close(fd);
	} while (e && (errno == EINTR));
}

/*
#include <iostream>
using namespace std;

int
main(int argc, char** argv)
{
	String out, err;
	int status;
	vector<String> arguments;

	String path;
	if (argc < 2) {
		cout << "enter path to execute: ";
		cin >> path;
	} else
		path = argv[1];

	for (int i=2; i<argc; i++)
		arguments.push_back(argv[i]);

	cout << "executing " << path << endl;
	cout << execute(path, arguments, out, err, status) << endl;

	cout << "stdout:" << endl << out << endl << endl;
	cout << "stderr:" << endl << err << endl << endl;
	cout << "status: " << status << endl;
}
*/
